// hard disk init routine 

#include "hsamio.h"
#include "hdstr.h"
#include "hmalloc.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>


// global vars with initalisers (stored in 'rom' or external to program)
volatile char scrnsaver @ 0x5A32;
volatile port uchar hddport @ 189;
volatile port uchar hdrport @ 191;

// vars with standard types
uchar  writeverify = 1; // set to 1 to verify writes;
uchar  portdata;
uint   hdmaxtrk;
uchar  hdmaxhead;
uchar  hdmaxsect;
uint   badcount;

// vars of structures
s_label  *mylabel;
s_dir    *mydir;
s_fsam   *myfsam;
s_fsam   *badfsam;

// procedure declarations
void waitforrdy();
void hdinit();
int  hdread(void *hddata, long se);
int  hdwrite(void *hddata, long se);
int  hdchkrdy();
void hdsetup(long se);
void rdbuf(void *buf);
void wrbuf(void *buf);
extern void exit(int);

// the code
void waitforrdy()
{
  do
  { hdrport = 0xF7;
    portdata=hddport;
    portdata &= 0x80;
  } while (portdata);
}


void hdinit()
{
 uint dummy; //,spt,hpc;
// char *buffer,*ptr;

 //   hdrport = 0xDF;	// send hardware reset
 //   dummy   = hddport;
 //   pause(20);
  
  hdrport = 0xEE;	// send software reset	
  hddport = 0x0E;
  hddport = 0x0;
  pause(20);

  pause(5);
  hddport = 0xA;	// clear software reset
  hddport = 0x0;
  pause(20);

  hdrport = 0x0FF;	// clear control reg on interface
  dummy   = hddport;
  dummy   = hddport;
  pause(20);

  waitforrdy();

}


int hdread(void *hddata, long se)
// read sector se into buffer hddata
// hddata must be at least 512 bytes long
{
 hdsetup(se);
 hdrport = 0xF7;
 hddport = 0x20;
 hddport = 0x0;
 waitforrdy();
 hdrport = 0xF0;
 rdbuf(hddata);
 return(hdchkrdy());
}


int hdwrite(void *hddata, long se)
// write sector se with data from hddata
{
 hdsetup(se);
 hdrport = 0xF7;
 hddport = 0x31;
 hddport = 0x0;
 waitforrdy();
 hdrport = 0xF0;
 wrbuf(hddata);
 waitforrdy();
 hdsetup(se);
 hdrport = 0xF7;
 hddport = 0x41;
 hddport = 0x0;
 waitforrdy();
 return(hdchkrdy());
}
 

int hdchkrdy()
{
 uint status,dummy;

 hdrport = 0xF7;
 status = hddport;
 dummy  = hddport;
// if ((status == 80) || (status == 82)) return(ok);
 status &= 0x21;
 if (!status) return(ok);
 return(error);
}


void hdsetup(long se)
{
 uchar trhi,trlo,head,sect;

 sect = se % hdmaxsect;
 se   /=     hdmaxsect;
 head = se % hdmaxhead;
 se   /=     hdmaxhead;
 trlo = se % 256;
 trhi = se / 256;

 sect++;
 head &= 0x1F;
 head |= 0xA0;

 hdrport = 243;
 hddport = sect;
 hddport = 0;

 hdrport = 244;
 hddport = trlo;
 hddport = 0;

 hdrport = 245;
 hddport = trhi;
 hddport = 0;

 hdrport = 246;
 hddport = head;
 hddport = 0;

 hdrport = 242;
 hddport = 1;
 hddport = 0;

}


void rdbuf(void *buf)
{ // buf passed in DE reg
#asm
 EX DE,HL
 di
 LD BC,189
 INIR
 INIR
 ei
#endasm
}


void wrbuf(void *buf)
{ // buf passed in DE reg
#asm
 EX DE,HL
 di
 LD BC,189
 OTIR
 OTIR
 ei
#endasm
}

int hdintread (unsigned int sect,unsigned int head,
               unsigned int track)
{
     unsigned int status,dummy;


     hdrport = 243;
     hddport = sect;
     hddport = 0;

     hdrport = 244;
     hddport = track % 256;
     hddport = 0;

     hdrport = 245;
     hddport = track / 256;
     hddport = 0;

     hdrport = 246;
     head   &= 0x1F;
     head   |= 0xA0;    // use CMA mode
     hddport = head;
     hddport = 0;

     hdrport = 242;
     hddport = 1;
     hddport = 0;


     hdrport = 0xF7; // sending command
     hddport = 0x21; // read
     hddport = 0x0;  //   "      "
     waitforrdy();
     hdrport = 0xF0; // reading data
     rdbuf(mydir);
     hdrport = 0xf7;
     status = hddport;
     dummy  = hddport;

     status &= 0x21;   // test if error bit set
     if (!status) return(ok);	
     return (error);
     
}

void getdisksize()
// find out the number of sectors and heads by looking
// at the drive 
{
 unsigned int sec,hed,ook,tmp,trk;
 
 sec = 0;
 hed = 0;
 trk = 0;
 
 do 
 { /* find number of sectors  */
    sec++ ;
    ook = hdintread(sec,0,0);
 } while (ook==ok);
 sec-- ;
 
 do 
 { /* find number of heads   */
    hed++ ;
    ook = hdintread(sec,hed,0);
    if (hed > 0xF) ook=error;
 } while (ook==ok);
 hed-- ;
 
 tmp = 0x8000;
 trk = 0;
 ook = 1;
 do 
 { /* find number of tracks */
    if (ook==ok)
       trk += tmp;
    else
       trk -= tmp;
    ook = hdintread(sec,hed,trk);
    tmp /= 2;
 } while (tmp);
 trk--;
 
 hdmaxtrk   = trk+1;
 hdmaxhead  = hed+1;
 hdmaxsect  = sec;

}

void markbad(long se)
{
// mark the passed sector as bad and build a bad sector table
badcount++;
   if (se < 10L)
   {
	   printf("\rDirectory area bad.");
	   printf("\rDisc is unuseable.");
	   exit(0);
   }
   if (badfsam->f_runen[badfsam->f_run] == 0)
   {
	   badfsam->f_runen[badfsam->f_run] = se;
	   badfsam->f_runst[badfsam->f_run] = se;
	   return;
   }
   if (badfsam->f_runst[badfsam->f_run] == (se+1))
   {
	   badfsam->f_runst[badfsam->f_run] = se;
	   return;
   }
   if (++badfsam->f_run >= n_run-1)
   {
	   printf("\rToo many bad arears.");
	   printf("\rFormat aborted.");
	   exit(0);
   }
   markbad(se);
}

void acceptsize()
// get heads/tracks/sectors from user
{
	int inint;
	
	printf("\r\rEnter number of tracks ");
	scanf("%d",&inint); hdmaxtrk = inint;
	printf("\rEnter number of heads ");
	scanf("%d",&inint); hdmaxhead = inint;
	printf("\rEnter number of sectors ");
	scanf("%d",&inint); hdmaxsect = inint;

}
	
main()
{
 char chin,manual; 
 long se,maxsect;

  manual = chin = 'n';
   
  do
 	{
		 clrscr();
	
		 printf("\rIDE Disc format routine v 3.0\r\r");
		 printf("Sizing disc. Please wait.\r");
		 hdinit();
		 
		 if (manual != 'y') getdisksize();
		 
		 printf("Drive parameters\r");
		 printf("number of cylinders  = %d\r",hdmaxtrk);
		 printf("heads per cylinder   = %d\r",hdmaxhead);
		 printf("sectors per track    = %d\r",hdmaxsect);
		 se  = (long)hdmaxsect;
		 se *= (long)hdmaxhead;
		 se *= (long)hdmaxtrk;
		 maxsect = se - 1;
		 se /= 2L;
		 printf("total disc capacity  = %ldK bytes\r\r",se);
		 badcount = 0;

		 printf("Accept these parameters (y/n) ");
		 while ((chin=tolower(getch())) == 0 );
		 if (chin != 'y')
		 {
			 manual = 'y';
			 acceptsize();
		 }

	} while (chin !='y');
	
 printf("\rDo surface test? (Y/N)  ");
 while ((chin=toupper(getch())) == 0 );
 if (chin != '!') if (chin != 'Y') chin = 'N';
 printf("%c\r",chin);
 writeverify = 1;

 if ((badfsam = calloc(1,sizeof(s_fsam))) == NULL)
 {
	printf("\rCould not allocate memory for bad sector table\r");
	printf("\rFormat aborted\r");
	exit(0);
 }

 if ((mydir = calloc(1,sizeof(s_dir))) == NULL)
 {
  printf("\rcould not allocate memory for disc buffer.");
  return;
 }
 scrnsaver = 1;  // keep screen active
     
 if ((chin == 'Y') || (chin == '!'))
 {
  /* write /0xef to all bytes in all sectors */
  
  memset(mydir,0xEF,(sizeof(s_dir)));
  if (chin == '!') maxsect = 1000;
  se = maxsect;
  for (;se>=0;se--)
  {
   printf("sector %ld %c\r\13",se,chin);
   chin = ' ';
   if (hdwrite(mydir,se)==error) 
   {
	   markbad(se);
	   chin='b';
   }
  }
   
 }
 
// write disk label sector 0
// printf("\r\rwriting label ");
 mylabel = (s_label*)mydir;
 memset(mydir,0,sizeof(s_dir));
 memcpy(mylabel->l_name,"SAM Hard drive v1.0\0",20);
 mylabel->l_sects  = hdmaxsect;
 mylabel->l_heads  = hdmaxhead;
 mylabel->l_tracks = hdmaxtrk;
// mylabel->l_cre.t_YY = 0;
// mylabel->l_cre.t_MM = 0;
// mylabel->l_cre.t_DD = 0;
// mylabel->l_cre.t_hh = 0;
// mylabel->l_cre.t_mm = 0;
// mylabel->l_cre.t_ss = 0;
 mylabel->l_free = 1L;
 mylabel->l_dir = 2L;
 if (hdwrite(mydir,0L)==error) markbad(0L);
 
// write free chain fsam sector 1
// printf("\rcreating free chain");
 memset(mydir,0,sizeof(s_dir));
 myfsam=(s_fsam*)mydir;
 myfsam->f_myadr   = 1L;
// myfsam->f_next    = 0L;
// myfsam->f_prev    = 0L;
// myfsam->f_dirty   = 0;
 myfsam->f_runst[0]= 6L;
 myfsam->f_runen[0]= maxsect;
 if (badcount)
 {
	 uchar idx;
	 for (idx = 0;badfsam->f_runst[idx];idx++)
	 {
		 myfsam->f_runen[idx]   = badfsam->f_runst[idx] -1;
		 myfsam->f_runst[idx+1] = badfsam->f_runen[idx] +1;
		 myfsam->f_runen[idx+1] = maxsect;
	 }
 }
 myfsam->f_fib.f_type=ft_Hfree;
 memcpy(myfsam->f_fib.f_name,"Free_space",10);
 if (hdwrite(mydir,1L)==error) markbad(1L);
 
// write directory sector 2
// printf("\rwriting directory");
 mydir = (s_dir*)mydir;
 memset(mydir,0,sizeof(s_dir));
 memcpy(mydir->d_head[0].h_name,".         ",namelen);
 mydir->d_this = 2L;
// mydir->d_next = 0L;
// mydir->d_dirty=0;
// mydir->d_head[0].h_dir  = 0;
 mydir->d_head[0].h_fsam = 3L;
 memcpy(mydir->d_head[1].h_name,"..        ",namelen);
// mydir->d_head[1].h_dir  = 0;  
 mydir->d_head[1].h_fsam = 4L;
 
 mydir->d_head[2].h_dir = baddir;
 memcpy(mydir->d_head[2].h_name,"Bad Bits  ",namelen);
 mydir->d_head[2].h_fsam = 5L;	 

 if (hdwrite(mydir,2L)==error) markbad(2L);
 
// write fsam for . sector 3
 memset(mydir,0,sizeof(s_dir));
 myfsam->f_fib.f_type = ft_Hdir | ft_hidden | ft_protected;
 memcpy(myfsam->f_fib.f_name,".         ",namelen); // my directory
 myfsam->f_myadr   = 3L;
// myfsam->f_next    = 0L;
// myfsam->f_prev    = 0L;
// myfsam->f_dirty   = 0;
// myfsam->f_fib.f_info.u_dirinfo.dirnum = 0;
 if (hdwrite(mydir,3L)==error) markbad(3L);
     
// write fsam for .. sector 4
 memset(mydir,0,sizeof(s_dir));
 myfsam->f_fib.f_type = ft_Hdir | ft_hidden | ft_protected;
 memcpy(myfsam->f_fib.f_name,"..        ",namelen); // my parents directory
 myfsam->f_myadr   = 4L;
// myfsam->f_next    = 0L;
// myfsam->f_prev    = 0L;
// myfsam->f_dirty   = 0;
// myfsam->f_fib.f_info.u_dirinfo.dirnum = 0;
 if (hdwrite(mydir,4L)==error) markbad(4L);

// write fsam for bad bits sector 5
 badfsam->f_fib.f_type = ft_Hfree | ft_hidden | ft_protected;
 memcpy(badfsam->f_fib.f_name,"Bad Bits  ",namelen);
 badfsam->f_myadr = 5L;
// badfsam->f_next    = 0L;
// badfsam->f_prev    = 0L;
// badfsam->f_dirty   = 0;
 badfsam->f_fib.f_info.u_dirinfo.dirnum = baddir;
 if (hdwrite(badfsam,5L)==error) markbad(5L);

  
 free(mydir);   
 free(badfsam);
 
 printf("\rFormat Completed %d bad sectors",badcount);
 
}
